#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "rle.h"
#include "huffman.h"
#include "lz.h"
#include "shannonfano.h"
#include "minilzo.h"

#define IN_SIZE 10000

// Pro LZO
#if defined(__LZO_STRICT_16BIT)
#define IN_LEN      (8*1024u)
#elif defined(LZO_ARCH_I086) && !defined(LZO_HAVE_MM_HUGE_ARRAY)
#define IN_LEN      (60*1024u)
#else
#define IN_LEN      (128*1024ul)
#endif
#define OUT_LEN     (IN_LEN + IN_LEN / 16 + 64 + 3)

static unsigned char __LZO_MMODEL in  [ IN_LEN ];
static unsigned char __LZO_MMODEL out [ OUT_LEN ];

#define HEAP_ALLOC(var,size) \
	    lzo_align_t __LZO_MMODEL var [ ((size) + (sizeof(lzo_align_t) - 1)) / sizeof(lzo_align_t) ]

static HEAP_ALLOC(wrkmem,LZO1X_1_MEM_COMPRESS);
// Pro LZO konec

void vypis(struct timespec cas1s, struct timespec cas2s,
		long in_len, long out_len, char *komprese, FILE *fo)
{
	double cas1, cas2, cas, in, out, ratio, speed;
	cas1 = (double) cas1s.tv_sec + (double) cas1s.tv_nsec / 1000000000;
        cas2 = (double) cas2s.tv_sec + (double) cas2s.tv_nsec / 1000000000;
	cas = cas2 - cas1;
	in = (double) in_len / 1000;
	out = (double) out_len / 1000;
	if (in_len != 0) ratio = (double) out_len / (double) in_len * 100;
	else ratio = 0;
	if (cas != 0) speed = (double) in_len / 1000 / cas;
	else speed = 0;
	printf("%s\n", komprese);
	printf("\t    t  %0.6lf s\n", cas);
	printf("\t   in  %0.3f kB\n", in);
	printf("\t  out  %0.3f kB\n", out);
	printf("\tratio  %0.1lf %\n", ratio);
	printf("\tspeed  %0.3f kB/s\n\n", speed);
	fprintf(fo, "%s\t%0.6lf\t%0.3f\t%0.3f\t%0.1lf\t%0.0f\n",
			komprese, cas, in, out, ratio, speed);
}

int main(int argc, char **argv)
{
	FILE *fr, *fw, *fo;
	char *data, *compressed;
	long in_len, out_len;
	unsigned int *work;
	struct timespec cas1s;
	struct timespec cas2s;
	double cas1, cas2;

// Osetreni nezadanych parametru
	if (argc < 3)
	{
		printf("Spuštění programu:\n%s in.dat log.txt\n", argv[0]);
		return 1;
	}

// Otevreni a inicializace souboru namerenych parametru
	fo = fopen(argv[2], "w");
	if (fo == NULL)
	{
		printf("Nemohu vytvorit log soubor. Overte prava zapisu.\n");
		return 1;
	}
	fprintf(fo, "Algoritmus\tCas [s]\tVstup [kB]\tVystup [kB]\tPomer [%]\tRychlost [kB/s]\n");

// Zjisteni velikosti vstupniho souboru
	fr = fopen(argv[1], "rb");
	if (fr == NULL)
	{
		printf("Zadany vstupni soubor nenalezen, nebo se jej nepodarilo otevrit.\n");
		return 1;
	}
	fseek(fr, 0L, SEEK_END);
	in_len = ftell(fr);
	fseek(fr, 0L, SEEK_SET);

// Nacteni obsahu souboru do pameti,
	data = malloc(in_len);
	in_len = fread(data, sizeof(char), in_len, fr);
	fclose(fr);

// Alokovani pameti pro kompresi dle nejnarocnejsiho algoritmu
	compressed = malloc(in_len + in_len / 16 + 384);

// RLE
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas1s);
	out_len = RLE_Compress(data,compressed,in_len);
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas2s);
	vypis(cas1s, cas2s, in_len, out_len, "RLE", fo);
	
	fw = fopen("CompressedRLE.dat", "wb");
	if (fw == NULL)
	{
		printf("Nelze ulozit komprimovana data do souboru.\n");
		return 1;
	}
	fwrite(compressed, sizeof(char), out_len, fw);
	fclose(fw);

// Huffman
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas1s);
	out_len = Huffman_Compress(data,compressed,in_len);
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas2s);
	vypis(cas1s, cas2s, in_len, out_len, "Huffman", fo);

	fw = fopen("CompressedHuffman.dat", "wb");
	if (fw == NULL)
        {
                printf("Nelze ulozit komprimovana data do souboru.\n");
                return 1;
        }
	fwrite(compressed, sizeof(char), out_len, fw);
	fclose(fw);

if (argc != 4)
{
// Lempel-Ziv
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas1s);
	out_len = LZ_Compress(data,compressed,in_len);
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas2s);
        vypis(cas1s, cas2s, in_len, out_len, "LZ", fo);

	fw = fopen("CompressedLz.dat", "wb");
	if (fw == NULL)
	{
		printf("Nelze ulozit komprimovana data do souboru.\n");
                return 1;
        }
	fwrite(compressed, sizeof(char), out_len, fw);
	fclose(fw);
}

// Lempel-Ziv-Fast
	work = malloc(sizeof(unsigned int) * (65536 + in_len));
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas1s);
	out_len = LZ_CompressFast(data,compressed,in_len,work);
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas2s);
	vypis(cas1s, cas2s, in_len, out_len, "LZ-Fast", fo);
	
	fw = fopen("CompressedLzFast.dat", "wb");
	if (fw == NULL)
        {
                printf("Nelze ulozit komprimovana data do souboru.\n");
                return 1;
        }
	fwrite(compressed, sizeof(char), out_len, fw);
	fclose(fw);
                                                         	
// Shannon-Fano
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas1s);
	out_len = SF_Compress(data,compressed,in_len);
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas2s);
        vypis(cas1s, cas2s, in_len, out_len, "SF", fo);

	fw = fopen("CompressedShannonfano.dat", "wb");
	if (fw == NULL)
	{
                printf("Nelze ulozit komprimovana data do souboru.\n");
                return 1;
        }
	fwrite(compressed, sizeof(char), out_len, fw);
	fclose(fw);

// MiniLZO
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas1s);
	lzo_init();
	lzo1x_1_compress(data,in_len,compressed,&out_len,wrkmem);
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cas2s);
        vypis(cas1s, cas2s, in_len, out_len, "MiniLZO", fo);

	fw = fopen("CompressedMiniLZO.dat", "wb");
	if (fw == NULL)
        {
                printf("Nelze ulozit komprimovana data do souboru.\n");
                return 1;
        }
	fwrite(compressed, sizeof(char), (long) out_len, fw);
	fclose(fw);

	fclose(fo);

	return 0;
}

